"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts_simple_ast_1 = require("ts-simple-ast");
const typescript_ast_util_1 = require("typescript-ast-util");
const util_1 = require("../util");
const hrtime_now_1 = require("hrtime-now");
/**

# description

function-like is returning a value but no return type was declared in signature. suggest adding a return type inferring from return value

# example

```const other = (a: string): number => {}```

# Attacks

"code": "2355",	"message": "A function whose declared type is neither 'void' nor 'any' must return a value.",

*/
exports.addReturnStatement = {
    name: 'addReturnStatement',
    config: {},
    predicate: (arg) => {
        const kind = typescript_ast_util_1.getKindName(arg.containingTargetLight);
        const kindToIncludeAnyOf = ['Identifier', 'Keyword', 'Type'];
        if (kindToIncludeAnyOf.find(k => kind.includes(k)) &&
            arg.diagnostics.find(d => d.code === 2355 && d.start === arg.containingTargetLight.getStart())) {
            return true;
        }
        else {
            arg.log(`predicate false because code is not 2355 or child.kind===${kind} dont include any of ${kindToIncludeAnyOf.join(', ')}`);
            return false;
        }
    },
    description: (arg) => `Add return statement`,
    apply: (arg) => {
        const t0 = hrtime_now_1.now();
        // try to get the type so we can return a nice object instead of just null
        const fn = arg.simpleNode.getAncestors().find(ts_simple_ast_1.TypeGuards.isSignaturedDeclaration);
        let returnValueString = `null`;
        if (fn && fn.getReturnType()) {
            returnValueString = util_1.getDefaultValueForType(fn.getReturnType());
        }
        const firstStatementedNode = arg.simpleNode.getAncestors().find(ts_simple_ast_1.TypeGuards.isStatementedNode);
        if (firstStatementedNode) {
            firstStatementedNode.addStatements(`return ${returnValueString};`);
        }
        else {
            arg.log('apply aborted because firstStatementedNode is null');
        }
        arg.log(`apply took ${hrtime_now_1.timeFrom(t0)}`);
    }
};
//# sourceMappingURL=addReturnStatement.js.map