"use strict";
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
const hrtime_now_1 = require("hrtime-now");
const ts_simple_ast_extra_1 = require("ts-simple-ast-extra");
const typescript_ast_util_1 = require("typescript-ast-util");
const typescript_plugin_util_1 = require("typescript-plugin-util");
const ts_module = __importStar(require("typescript/lib/tsserverlibrary"));
const codeFixes_1 = require("./codeFixes");
const PLUGIN_NAME = 'typescript-plugin-proactive-code-fixes';
const REFACTOR_ACTION_NAME = `${PLUGIN_NAME}-refactor-action`;
let target;
function getApplicableRefactors(fileName, positionOrRange, userPreferences) {
    const refactors = info.languageService.getApplicableRefactors(fileName, positionOrRange, userPreferences) || [];
    const codeFix = getCodeFix(fileName, positionOrRange);
    if (!codeFix) {
        return refactors;
    }
    target = codeFix.target;
    refactors.push({
        name: `${PLUGIN_NAME}-refactor-info`,
        description: 'Code Fixes',
        actions: codeFix.fixes.map(fix => ({
            name: REFACTOR_ACTION_NAME + '-' + fix.name,
            description: fix.description(target)
        }))
    });
    return refactors;
}
function getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, userPreferences) {
    const t0 = hrtime_now_1.now();
    log(`getEditsForRefactor ${positionOrRange} ${refactorName} ${actionName}`);
    const refactors = info.languageService.getEditsForRefactor(fileName, formatOptions, positionOrRange, refactorName, actionName, userPreferences);
    if (!actionName.startsWith(REFACTOR_ACTION_NAME) || !target.containingTarget) {
        return refactors;
    }
    const fixName = actionName.substring(REFACTOR_ACTION_NAME.length + 1, actionName.length);
    const fix = codeFixes_1.codeFixes.find(fix => fix.name === fixName);
    if (!fix) {
        log(`no getEditsForRefactor because no fix was found for actionName == ${actionName}`);
        return refactors;
    }
    const result = applyCodeFix(fix, target, formatOptions, positionOrRange, userPreferences);
    log(`getEditsForRefactor total time took ${hrtime_now_1.timeFrom(t0)}`);
    if (typeof (result) === 'string') {
        return {
            edits: [
                {
                    fileName, textChanges: [
                        {
                            newText: result,
                            span: {
                                start: 0,
                                length: target.sourceFile.getText().length
                            }
                        }
                    ]
                }
            ]
        };
    }
    if (result && result.edits) {
        return result;
    }
    return refactors;
}
function getCodeFix(fileName, positionOrRange, end, errorCodes, formatOptions) {
    const t0 = hrtime_now_1.now();
    const program = info.languageService.getProgram();
    const sourceFile = program.getSourceFile(fileName);
    if (!sourceFile) {
        log(`getCodeFix false because !sourceFile`);
        return;
    }
    const getDiagnosticT0 = hrtime_now_1.now();
    let diagnostics;
    const start = typescript_ast_util_1.positionOrRangeToNumber(positionOrRange);
    if (errorCodes) {
        diagnostics = errorCodes.map(code => ({
            file: sourceFile,
            start: start,
            length: end - start,
            messageText: 'dummy',
            category: ts_module.DiagnosticCategory.Error,
            code
        }));
    }
    else {
        diagnostics = ts.getPreEmitDiagnostics(program, sourceFile);
    }
    log(`getPreEmitDiagnostics took ${hrtime_now_1.timeFrom(getDiagnosticT0)}`);
    const range = typescript_ast_util_1.positionOrRangeToRange(start + 1);
    const containingTarget = typescript_ast_util_1.findChildContainingRange(sourceFile, range) || sourceFile;
    const containingTargetLight = typescript_ast_util_1.findChildContainingRangeLight(sourceFile, range) || sourceFile;
    const containedTarget = typescript_ast_util_1.findChildContainedRange(sourceFile, range) || sourceFile;
    const codeFixesFilterT0 = hrtime_now_1.now();
    const target = { diagnostics, containingTarget, containingTargetLight, containedTarget, log, program, sourceFile, positionOrRange };
    const fixes = codeFixes_1.codeFixes.filter(fix => {
        try {
            return fix.predicate(target);
        }
        catch (ex) {
            log('getCodeFix exception in plugin predicate ' + fix.name + ex + '\n' + ex.stack);
        }
    });
    log(`codeFixesFilterT0 took ${hrtime_now_1.timeFrom(codeFixesFilterT0)}`);
    if (!fixes || !fixes.length) {
        log(`no getCodeFix because no codeFixes returned true for node of kind ==${typescript_ast_util_1.getKindName(containingTarget.kind)} and diagnostics: ${diagnostics.map(d => d.code)}`);
        return;
    }
    log(`getCodeFix took ${hrtime_now_1.timeFrom(t0)}`);
    return { fixes, target };
}
let currentFix;
function applyCodeFix(fix, options, formatOptions, positionOrRange, userPreferences) {
    let simpleProject;
    let sourceFile;
    const fileName = options.sourceFile.fileName;
    if (fix && fix.needSimpleAst !== false) {
        const createSimpleASTProjectT0 = hrtime_now_1.now();
        simpleProject = ts_simple_ast_extra_1.getSimpleProject(info.project, formatOptions, userPreferences);
        log(`applyCodeFix createSimpleASTProject took ${hrtime_now_1.timeFrom(createSimpleASTProjectT0)}`);
        const simpleNodeT0 = hrtime_now_1.now();
        sourceFile = simpleProject.getSourceFile(options.sourceFile.fileName);
        // update sourceFile (read from FS) with current text in editor buffer (since it could have non saved changes)
        const textInBuffer = info.languageService.getProgram().getSourceFile(sourceFile.getFilePath()).getText();
        sourceFile.replaceWithText(textInBuffer);
        options.simpleNode = sourceFile.getDescendantAtPos(typescript_ast_util_1.positionOrRangeToNumber(positionOrRange)) || sourceFile;
        options.simpleProject = simpleProject;
        if (!options.simpleNode) {
            log(`no applyCodeFix because sourceFile is null for fileName=== ${fileName}`);
            return null;
        }
        log(`applyCodeFix first getSourceFile() and simpleNode took ${hrtime_now_1.timeFrom(simpleNodeT0)} and node.kind is ${options.simpleNode.getKindName()}`);
    }
    // we are ready, with or without ast-simple to perform the change
    const fixapplyT0 = hrtime_now_1.now();
    currentFix = fix;
    let result;
    try {
        result = fix.apply(options);
    }
    catch (error) {
        log(`applyCodeFix fix.apply() Exception ${error} \n ${error.stack}`);
    }
    log(`applyCodeFix fix.apply() took ${hrtime_now_1.timeFrom(fixapplyT0)}`);
    if (!result) {
        return sourceFile.getText();
    }
    return result;
}
let ts;
let info;
let log;
const pluginDefinition = { getApplicableRefactors, getEditsForRefactor };
module.exports = typescript_plugin_util_1.getPluginCreate(pluginDefinition, (modules, anInfo) => {
    ts = modules.typescript;
    info = anInfo;
    log = function (msg) {
        info.project.projectService.logger.info(`Plugin ${PLUGIN_NAME}, Fix: ${currentFix && currentFix.name}, Message: ${msg}`);
    };
    info.project.projectService.logger.info(`${PLUGIN_NAME} created`);
});
//# sourceMappingURL=index.js.map