"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const ts_simple_ast_1 = require("ts-simple-ast");
exports.buildParameterStructure = (p) => ({
    name: p.getNameOrThrow(),
    hasQuestionToken: p.hasQuestionToken(),
    type: p.getTypeNode() == null ? undefined : p.getTypeNodeOrThrow().getText(),
    isRestParameter: p.isRestParameter(),
    scope: p.hasScopeKeyword() ? p.getScope() : undefined
});
/** fix given function-like declaration parameters and type to implement given signature */
function fixSignature(decl, signature) {
    decl.setReturnType(signature.getReturnType().getText());
    // add missing params and fix exiting param types
    let memberParams = decl.getParameters();
    let signatureParams = signature.getParameters();
    for (let i = 0; i < signatureParams.length; i++) {
        const signatureParam = signatureParams[i];
        if (memberParams.length <= i) {
            decl.addParameter(exports.buildParameterStructure(signatureParam));
        }
        else {
            const memberParam = memberParams[i];
            if (!areTypesEqual(memberParam.getType(), signatureParam.getType())) {
                memberParam.set(Object.assign({}, exports.buildParameterStructure(signatureParam), { name: memberParam.getName() }));
            }
        }
    }
    // remove extra non optional params member signature might have
    memberParams = decl.getParameters();
    signatureParams = signature.getParameters();
    if (memberParams.length > signatureParams.length) {
        for (let i = signatureParams.length; i < memberParams.length; i++) {
            if (!(memberParams[i].hasInitializer() || memberParams[i].isOptional())) {
                memberParams[i].remove();
            }
        }
    }
}
exports.fixSignature = fixSignature;
/** dirty way of checking if two types are compatible */
function areTypesEqual(t1, t2) { return t1.getText().replace(/\s+/gi, '') === t2.getText().replace(/\s+/gi, ''); }
exports.areTypesEqual = areTypesEqual;
/**
 * returns all implements clauses of this class and its super classes both things recursively
 */
exports.getImplementsAll = (cl) => {
    let result = [];
    cl.getImplements().forEach(impl => {
        // TODO: types like A|B
        result.push(impl);
        impl.getType().getSymbolOrThrow().getDeclarations().forEach(d => {
            if (ts_simple_ast_1.TypeGuards.isInterfaceDeclaration(d)) {
                result = result.concat(exports.getExtendsRecursively(d));
            }
        });
    });
    exports.getExtendsRecursively(cl).forEach(ext => {
        ext.getType().getSymbolOrThrow().getDeclarations().forEach(d => {
            if (ts_simple_ast_1.TypeGuards.isClassDeclaration(d)) {
                result = result.concat(exports.getImplementsAll(d));
            }
        });
    });
    return result;
};
exports.getExtendsRecursively = (decl) => {
    let extendExpressions = ts_simple_ast_1.TypeGuards.isClassDeclaration(decl) ? (decl.getExtends() ? [decl.getExtends()] : []) : decl.getExtends();
    extendExpressions.forEach(expr => {
        expr.getType().getSymbol().getDeclarations().forEach(d => {
            if (ts_simple_ast_1.TypeGuards.isInterfaceDeclaration(d) || ts_simple_ast_1.TypeGuards.isClassDeclaration(d)) {
                extendExpressions = extendExpressions.concat(exports.getExtendsRecursively(d));
            }
        });
    });
    return extendExpressions;
};
exports.findInterfacesWithPropertyNamed = (decl, memberName) => exports.getImplementsAll(decl)
    .map(expr => expr.getType().getSymbolOrThrow().getDeclarations())
    .reduce((a, v) => a.concat(v), [])
    .filter(ts_simple_ast_1.TypeGuards.isInterfaceDeclaration)
    .filter(d => d.getMembers().find(m => ts_simple_ast_1.TypeGuards.isPropertyNamedNode(m) && m.getName() === memberName))
    .filter((value, pos, arr) => arr.indexOf(value) === pos); // union
/**
 * returns a default representation of given type, for example `''` for `string`, `[]` for `Array`, etc.
 * TODO: this should support recursion in case it references another interface - we can recreate the whole thing recursively...
 */
function getDefaultValueForType(t) {
    if (!t) {
        return 'null';
    }
    else if (t.getText() === 'string') {
        return '\'\'';
    }
    else if (t.getText() === 'boolean') {
        return 'false';
    }
    else if (t.getText() === 'number') {
        return '0';
    }
    else if (t.getText() === 'Date') {
        return 'new Date()';
    }
    else if (t.getText() === 'RegExp') {
        return '//ig';
    }
    else if (t.isArray()) {
        return '[]';
    }
    else {
        return 'null';
    }
}
exports.getDefaultValueForType = getDefaultValueForType;
function getName(node, defaultName = 'unknown_name') {
    const a = node;
    let id;
    return a && a.getName && a.getName() || (id = node.getDescendantsOfKind(ts_simple_ast_1.SyntaxKind.Identifier)) && id.getText() || defaultName;
}
exports.getName = getName;
function quote(s, q) {
    const newLiteral = s.replace(new RegExp(`${q}`, 'gmi'), `\\${q}`);
    return q + newLiteral + q;
}
exports.quote = quote;
function changeQuoteChar(node, newQuoteChar) {
    const newText = quote(node.getLiteralText(), newQuoteChar);
    node.replaceWithText(newText);
}
exports.changeQuoteChar = changeQuoteChar;
function buildRefactorEditInfo(sourceFile, newText, start = 0, length = 0) {
    return {
        edits: [
            {
                fileName: sourceFile.fileName,
                textChanges: [
                    {
                        newText,
                        span: { start, length }
                    }
                ]
            }
        ]
    };
}
exports.buildRefactorEditInfo = buildRefactorEditInfo;
//# sourceMappingURL=util.js.map